@@ -10,7 +10,7 @@ with open(version_file, 'r') as f:
 
 __version__ = version
 __authors__ = "Ad de Roo, Emiliano Gelati, Peter Burek, Johan van der Knijff, Niko Wanders"
-__date__ = "27/02/2020"
-__copyright__ = "Copyright 2020, European Commission - Joint Research Centre"
-__maintainer__ = "Emiliano Gelati, Domenico Nappo, Valerio Lorini, Lorenzo Mentaschi, Ad de Roo"
+__date__ = "09/06/2022"
+__copyright__ = "Copyright 2019-2022, European Commission - Joint Research Centre"
+__maintainer__ = "Stefania Grimaldi, Cinzia Mazzetti, Carlo Russo, Damien Decremer, Corentin Carton De Wiart, Valerio Lorini, Ad de Roo"
 __status__ = "Operation"
