@@ -19,10 +19,11 @@ from __future__ import print_function, absolute_import
 import os
 import warnings
 
-from pcraster import report, ifthen, mapmaximum, catchmenttotal
+from pcraster import report, ifthen, catchmenttotal, mapmaximum
 
 from .zusatz import TimeoutputTimeseries
-from .add1 import decompress, valuecell, loadmap, compressArray, writenet
+from .add1 import decompress, valuecell, loadmap, compressArray#, writenet
+from .netcdf import writenet
 from .errors import LisfloodFileError, LisfloodWarning
 from .settings import inttodate, CDFFlags, LisSettings
 
@@ -55,16 +56,13 @@ class outputTssMap(object):
         binding = settings.binding
         report_time_serie_act = settings.report_timeseries
         binding['Catchments'] = self.var.Catchments
-        binding['1'] = None  # ?? :( ??
         # output for single column eg mapmaximum
         self.var.Tss = {}
 
         for tss in report_time_serie_act:
             where = report_time_serie_act[tss].where
             outpoints = binding[where]
-            if where == "1":
-                pass
-            elif where == "Catchments":
+            if where == "Catchments":
                 outpoints = decompress(outpoints)
             else:
                 coord = binding[where].split()  # could be gauges, sites, lakeSites etc.
@@ -72,12 +70,12 @@ class outputTssMap(object):
                     outpoints = valuecell(self.var.MaskMap, coord, outpoints)
                 else:
                     try:
-                        outpoints = loadmap(where,pcr=True)
-                        outpoints = ifthen(outpoints != 0,outpoints)
-                          # this is necessary if netcdf maps are loaded !! otherwise strange dis.tss
-                    except:
-                        msg = "Setting output points\n"
-                        raise LisfloodFileError(outpoints,msg)
+                        outpoints = loadmap(where, pcr=True)
+                        outpoints = ifthen(outpoints != 0, outpoints)
+                        # this is necessary if netcdf maps are loaded !! otherwise strange dis.tss
+                    except Exception as e:
+                        msg = "Setting output points\n {}".format(str(e))
+                        raise LisfloodFileError(outpoints, msg)
 
             if option['MonteCarlo']:
                 if os.path.exists(os.path.split(binding[tss])[0]):
@@ -149,15 +147,20 @@ class outputTssMap(object):
 
         for maps in report_maps_end.keys():
             # report end map filename
-
-            try:
+            if settings.mc_set:
+                # MonteCarlo model
                 where = os.path.join(str(self.var.currentSampleNumber()), binding[maps].split("/")[-1])
-            except:
+            else:
                 where = binding.get(maps)
             if not where:
                 continue
-            # Output path and name of report map
             what = 'self.var.' + report_maps_end[maps].output_var
+
+            try:
+                out_var = eval(what)
+            except:
+                continue
+
             if where not in checkifdouble:
                 checkifdouble.append(where)
                 # checks if saved at same place, if no: add to list
@@ -176,42 +179,38 @@ class outputTssMap(object):
                             # CM mod: write end map to netCDF file (single)
                             # CM ##########################
 
-                            try:
-                                writenet(0, eval(what), where, self.var.DtDay, maps, report_maps_end[maps].output_var,
-                                         report_maps_end[maps].unit, 'f4', reportStartDate,
-                                         self.var.currentTimeStep(), self.var.currentTimeStep())
-                            except Exception as e:
-                                print(str(e), 'END', what, where, self.var.DtDay, maps, report_maps_end[maps].output_var,
-                                      report_maps_end[maps].unit, 'f4', reportStartDate,
-                                      self.var.currentTimeStep(), self.var.currentTimeStep())
-                            ################################
+                            writenet(0, out_var, where, self.var.DtDay, maps, report_maps_end[maps].output_var,
+                                     report_maps_end[maps].unit, 'd', reportStartDate,
+                                     self.var.currentTimeStep(), self.var.currentTimeStep())
 
                         else:
-                            report(decompress(eval(what)), str(where))
+                            report(decompress(out_var), str(where))
                     else:
                         if option['writeNetcdfStack']:
                             
-                            try:
-                                writenet(0, eval(what), where, self.var.DtDay, maps, report_maps_end[maps].output_var,
-                                         report_maps_end[maps].unit, 'f4', reportStartDate,
-                                         self.var.currentTimeStep(), self.var.currentTimeStep())
-                            except Exception as e:
-                                print(str(e), 'END', what, where, self.var.DtDay, maps, report_maps_end[maps].output_var,
-                                      report_maps_end[maps].unit, 'f4', reportStartDate,
-                                      self.var.currentTimeStep(), self.var.currentTimeStep())
-                            ###########################
+                            writenet(0, out_var, where, self.var.DtDay, maps, report_maps_end[maps].output_var,
+                                     report_maps_end[maps].unit, 'd', reportStartDate,
+                                     self.var.currentTimeStep(), self.var.currentTimeStep())
                         else:
-                            self.var.report(decompress(eval(what)), str(where))
+                            self.var.report(decompress(out_var), str(where))
 
         # Report REPORTSTEPS maps
         for maps in report_maps_steps.keys():
             # report reportsteps maps
+            if settings.mc_set:
+                # MonteCarlo model
+                where = os.path.join(str(self.var.currentSampleNumber()), binding[maps].split("/")[-1])
+            else:
+                where = binding.get(maps)
+            if not where:
+                continue
+            what = 'self.var.' + report_maps_steps[maps].output_var
 
             try:
-                where = os.path.join(str(self.var.currentSampleNumber()), binding[maps].split("/")[-1])
+                out_var = eval(what)
             except:
-                where = binding[maps]
-            what = 'self.var.' + report_maps_steps[maps].output_var
+                continue
+
             if not(where in checkifdouble):
                 checkifdouble.append(where)
                 # checks if saved at same place, if no: add to list
@@ -232,8 +231,7 @@ class outputTssMap(object):
                             frequency = "annual"
                     except:
                         yearly = False
-
-                    if (monthly and self.var.monthend) or (yearly and self.var.yearend) or not (monthly or yearly):
+                    if (monthly and self.var.monthend) or (yearly and self.var.yearend):
                         # checks if a flag monthly or yearly exists
                         if option['writeNetcdfStack']:
                             # Get start date for reporting start step
@@ -243,29 +241,44 @@ class outputTssMap(object):
                             # get step number for last reporting step
                             reportStepEnd = self.var.ReportSteps[-1] - self.var.ReportSteps[0] + 1
                             cdfflags = CDFFlags.instance()
-                            try:
-                                writenet(cdfflags[flagcdf], eval(what), where, self.var.DtDay, maps,
-                                         report_maps_steps[maps].output_var, report_maps_steps[maps].unit, 'f4',
-                                         reportStartDate, reportStepStart, reportStepEnd, frequency)
-                            except Exception as e:
-                                print(" +----> ERR: {}".format(str(e)))
-                                print("REP flag:{} - {} {} {} {} {} {} {} {} {} {}".format(
-                                      cdfflags[flagcdf], what, where, self.var.DtDay, maps,
-                                      report_maps_steps[maps].output_var, report_maps_steps[maps].unit, 'f4',
-                                      reportStartDate, reportStepStart, reportStepEnd
-                                      ))
-
+                            writenet(cdfflags[flagcdf], out_var, where, self.var.DtDay, maps,
+                                     report_maps_steps[maps].output_var, report_maps_steps[maps].unit, 'd',
+                                     reportStartDate, reportStepStart, reportStepEnd, frequency)      
                         else:
                             self.var.report(decompress(eval(what)), str(where))
-
+                    if frequency == "all":
+                        # checks if a flag monthly or yearly exists
+                        if option['writeNetcdfStack']:
+                            # Get start date for reporting start step
+                            reportStartDate = inttodate(self.var.ReportSteps[0] - 1, self.var.CalendarDayStart)
+                            # get step number for first reporting step
+                            reportStepStart = 1
+                            # get step number for last reporting step
+                            reportStepEnd = self.var.ReportSteps[-1] - self.var.ReportSteps[0] + 1
+                            cdfflags = CDFFlags.instance()
+                            writenet(cdfflags[flagcdf], out_var, where, self.var.DtDay, maps,
+                                     report_maps_steps[maps].output_var, report_maps_steps[maps].unit, 'd',
+                                     reportStartDate, reportStepStart, reportStepEnd, frequency)      
+                        else:
+                            self.var.report(decompress(eval(what)), str(where))
+                            
         # Report ALL maps
         for maps in report_maps_all.keys():
             # report maps for all timesteps
-            try:
+            if settings.mc_set:
                 where = os.path.join(str(self.var.currentSampleNumber()), binding[maps].split("/")[-1])
-            except:
-                where = binding[maps]
+            else:
+                where = binding.get(maps)
+            if not where:
+                continue
+
             what = 'self.var.' + report_maps_all[maps].output_var
+
+            try:
+                out_var = eval(what)
+            except:
+                continue
+
             if where not in checkifdouble:
                 checkifdouble.append(where)
                 # checks if saved at same place, if no: add to list
@@ -288,8 +301,8 @@ class outputTssMap(object):
                         frequency = "annual"
                 except:
                     yearly = False
-
-                if (monthly and self.var.monthend) or (yearly and self.var.yearend) or not (monthly or yearly):
+                ## stef
+                if (monthly and self.var.monthend) or (yearly and self.var.yearend):
                     # checks if a flag monthly or yearly exists]
                     if option['writeNetcdfStack']:
                         #Get start date for reporting start step
@@ -301,17 +314,31 @@ class outputTssMap(object):
                         #get step number for last reporting step which is always the last simulation step
                         #last simulation step referred to reportStartDate
                         reportStepEnd = binding['StepEndInt'] - binding['StepStartInt'] + 1
+                        
+                        cdfflags = CDFFlags.instance()
+                        writenet(cdfflags[flagcdf], out_var, where, self.var.DtDay, maps, report_maps_all[maps].output_var,
+                                 report_maps_all[maps].unit, 'd', reportStartDate, reportStepStart, reportStepEnd, frequency)
 
-                        try:
-                            cdfflags = CDFFlags.instance()
-                            writenet(cdfflags[flagcdf], eval(what), where, self.var.DtDay, maps, report_maps_all[maps].output_var,
-                                     report_maps_all[maps].unit, 'f4', reportStartDate, reportStepStart, reportStepEnd, frequency)
-                        except Exception as e:
-                            warnings.warn(LisfloodWarning(str(e)))
-                            print(str(e), "ALL", what, where, self.var.DtDay, maps, report_maps_all[maps].output_var,
-                                  report_maps_all[maps].unit, 'f4', reportStartDate,reportStepStart, reportStepEnd)
                     else:
                         self.var.report(decompress(eval(what)), trimPCRasterOutputPath(where))
+                if frequency == "all":
+                    if option['writeNetcdfStack']:
+                        #Get start date for reporting start step
+                        reportStartDate = inttodate(binding['StepStartInt'] - 1, self.var.CalendarDayStart)
+                        # CM: get step number for first reporting step which is always the first simulation step
+                        # CM: first simulation step referred to reportStartDate
+                        ##reportStepStart = int(binding['StepStart'])
+                        reportStepStart = 1
+                        #get step number for last reporting step which is always the last simulation step
+                        #last simulation step referred to reportStartDate
+                        reportStepEnd = binding['StepEndInt'] - binding['StepStartInt'] + 1
+                        
+                        cdfflags = CDFFlags.instance()
+                        writenet(cdfflags[flagcdf], out_var, where, self.var.DtDay, maps, report_maps_all[maps].output_var,
+                                 report_maps_all[maps].unit, 'd', reportStartDate, reportStepStart, reportStepEnd, frequency)
+
+                    else:
+                        self.var.report(decompress(eval(what)), trimPCRasterOutputPath(where))                         
 
         cdfflags = CDFFlags.instance()
         # set the falg to indicate if a netcdffile has to be created or is only appended
