@@ -30,11 +30,13 @@ import os
 import pprint
 import inspect
 from collections import namedtuple
-
+import multiprocessing
+import cftime
+import threading
 import xml.dom.minidom
 import pcraster
 from netCDF4 import Dataset, date2num, num2date
-from pandas.core.tools.datetimes import parse_time_string
+from pandas.core.tools.datetimes import parsing
 import numpy as np
 
 from .errors import LisfloodError, LisfloodWarning, LisfloodFileError
@@ -79,6 +81,45 @@ class Singleton(type):
         return cls._current[cls]
 
 
+class ThreadSingleton(type):
+    """
+    Singleton metaclass to keep single instances by init arguments and by thread
+    """
+    _instances = {}
+    _current = {}
+
+    def __init__(cls, name, bases, dct):
+        super(ThreadSingleton, cls).__init__(name, bases, dct)
+
+    def __call__(cls, *args, **kwargs):
+        init = cls.__init__
+        task_id = cls.get_task_id()
+        if init is not None:
+            init_args = inspect.getcallargs(init, None, *args, **kwargs).items()
+            new_init_args = []
+            for a in init_args:
+                if isinstance(a[1], np.ndarray):
+                    new_init_args.append((a[0], a[1].tostring()))
+                else:
+                    new_init_args.append(a)
+            key = (task_id, cls, str(new_init_args))
+        else:
+            key = (task_id, cls)
+
+        if key not in cls._instances:
+            cls._instances[key] = super(ThreadSingleton, cls).__call__(*args, **kwargs)
+        cls._current[(task_id, cls)] = cls._instances[key]
+        return cls._instances[key]
+
+    def instance(cls):
+        task_id = cls.get_task_id()
+        key = (task_id, cls)
+        return cls._current[key]
+
+    def get_task_id(cls):
+        task_id = multiprocessing.current_process().name + '_' + str(threading.get_ident())
+        return task_id
+
 @nine
 class CDFFlags(with_metaclass(Singleton)):
     """
@@ -86,8 +127,8 @@ class CDFFlags(with_metaclass(Singleton)):
     flagcdf = 0
     flagcdf = 1  # index flag for writing nedcdf = 1 (=steps) -> indicated if a netcdf is created or maps are appended
     flagcdf = 2
-    flagcdf = 4  # set to yearly (step) flag
     flagcdf = 3  # set to monthly (step) flag
+    flagcdf = 4  # set to yearly (step) flag
     flagcdf = 5  # set to monthly flag
     flagcdf = 6  # set to yearly flag
     """
@@ -173,9 +214,28 @@ class NetCDFMetadata(with_metaclass(Singleton)):
             self.data[var] = {k: v for k, v in iteritems(nf1.variables[var].__dict__) if k != '_FillValue'}
         nf1.close()
 
+@nine
+class EPICSettings(with_metaclass(Singleton)):
+
+    def __init__(self):
+        self.soil_uses = ["Rainfed", 'Forest', 'Irrigated']
+        self.prescribed_vegetation = ['{}_prescribed'.format(fract) for fract in self.soil_uses]
+        self.vegetation_landuse = OrderedDict(
+            zip(self.prescribed_vegetation[:], self.soil_uses[:])
+        )
+        self.landuse_vegetation = OrderedDict(
+            [(v, [k]) for k, v in self.vegetation_landuse.items()]
+        )
+        self.prescribe_lai = OrderedDict(
+            zip(self.prescribed_vegetation[:], ['LAIOtherMaps', 'LAIForestMaps', 'LAIIrrigationMaps'])
+        )
+        self.landuse_inputmap = OrderedDict(
+                zip(self.landuse_vegetation.keys(), ["OtherFraction", "ForestFraction", "IrrigationFraction"])
+        )
+
 
 @nine
-class LisSettings(with_metaclass(Singleton)):
+class LisSettings(with_metaclass(ThreadSingleton)):
     printer = pprint.PrettyPrinter(indent=4, width=120)
 
     def __str__(self):
@@ -195,7 +255,7 @@ class LisSettings(with_metaclass(Singleton)):
                report_maps_end=self.printer.pformat(self.report_maps_end))
         return res
 
-    def __init__(self, settings_file):
+    def __init__(self, settings_file, sys_args=[]):
         dom = xml.dom.minidom.parse(settings_file)
         self.settings_dir = os.path.normpath(os.path.dirname((os.path.abspath(settings_file))))
         self.settings_path = os.path.normpath(os.path.abspath(settings_file))
@@ -206,24 +266,27 @@ class LisSettings(with_metaclass(Singleton)):
         self.ncores = self._ncores(user_settings)
         self.binding = bindings
         self.options = self._options(dom)
-        self.flags = self._flags()
+        self.flags = self._flags(sys_args)
         self.model_steps = self._model_steps()
         self.report_steps = self._report_steps(user_settings, bindings)
         self.filter_steps = self._filter_steps(user_settings)
         self.ens_members = self._ens_members(user_settings)
-        self.report_timeseries = self._report_tss()
-        self.report_maps_steps, self.report_maps_all, self.report_maps_end = self._reported_maps()
-        self.report_timeseries = {k: v for k, v in iteritems(self.report_timeseries) if v}
-        self.report_maps_steps = {k: v for k, v in iteritems(self.report_maps_steps) if v}
-        self.report_maps_all = {k: v for k, v in iteritems(self.report_maps_all) if v}
-        self.report_maps_end = {k: v for k, v in iteritems(self.report_maps_end) if v}
+        self.report_timeseries, self.report_maps_steps, self.report_maps_all, self.report_maps_end = self.build_reportedmaps_dicts()
         self.enkf_set, self.mc_set = self.montecarlo_kalman_settings()
         self.step_start, self.step_end = self.binding['StepStart'], self.binding['StepEnd']
         self.step_start_int, self.step_end_int = self.binding['StepStartInt'], self.binding['StepEndInt']
         ref_date_start = calendar(self.binding['CalendarDayStart'], self.binding['calendar_type'])
         self.step_start_dt = inttodate(self.step_start_int - 1, ref_date_start, binding=self.binding)
         self.step_end_dt = inttodate(self.step_end_int - 1, ref_date_start, binding=self.binding)
+        self.maskpath = self.binding['MaskMap']
 
+    def build_reportedmaps_dicts(self):
+        self.report_timeseries = self._report_tss()
+        self.report_maps_steps, self.report_maps_all, self.report_maps_end = self._reported_maps()
+        self.report_maps_steps = {k: v for k, v in iteritems(self.report_maps_steps) if v}
+        self.report_maps_all = {k: v for k, v in iteritems(self.report_maps_all) if v}
+        self.report_maps_end = {k: v for k, v in iteritems(self.report_maps_end) if v}
+        return self.report_timeseries, self.report_maps_steps, self.report_maps_all, self.report_maps_end
 
     def montecarlo_kalman_settings(self):
         # Ensemble Kalman filter
@@ -248,7 +311,7 @@ class LisSettings(with_metaclass(Singleton)):
             float(self.timestep_init)
         except ValueError:
             try:
-                parse_time_string(self.timestep_init, dayfirst=True)
+                parsing.parse_time_string(self.timestep_init, dayfirst=True)
             except ValueError:
                 raise LisfloodError('Option timestepInit was not parsable. Must be integer or date string: {}'.format(self.timestep_init))
             else:
@@ -325,18 +388,15 @@ class LisSettings(with_metaclass(Singleton)):
         return pathout
 
     @staticmethod
-    def _flags():
+    def _flags(sys_args):
         flags = OrderedDict([('quiet', False), ('veryquiet', False), ('loud', False),
                              ('checkfiles', False), ('noheader', False), ('printtime', False),
-                             ('debug', False), ('nancheck', False)])
-        if 'test' in sys.argv[0] or 'test' in sys.argv[1]:
-            # return defaults during tests
-            return flags
+                             ('debug', False), ('nancheck', False), ('initonly', False)])
 
         @cached
         def _flags(argz):
             try:
-                opts, arguments = getopt.getopt(argz, 'qvlchtdn', list(flags.keys()))
+                opts, arguments = getopt.getopt(argz, 'qvlchtdni', list(flags.keys()))
             except getopt.GetoptError:
                 from ..main import usage
                 usage()
@@ -344,14 +404,15 @@ class LisSettings(with_metaclass(Singleton)):
                 for o, a in opts:
                     for opt in (('-q', '--quiet'), ('-v', '--veryquiet'),
                                 ('-l', '--loud'), ('-c', '--checkfiles'),
-                                ('-h', '--noheader'), ('-t', '--printtime')):
+                                ('-h', '--noheader'), ('-t', '--printtime'),
+                                ('-d', '--debug'), ('-n', '--nancheck'),
+                                ('-i', '--initonly')):
                         if o in opt:
                             flags[opt[1].lstrip('--')] = True
                             break
             return flags
 
-        args = sys.argv[2:]
-        return _flags(args)
+        return _flags(sys_args)
 
     def _bindings(self, dom):
         binding = {}
@@ -385,12 +446,9 @@ class LisSettings(with_metaclass(Singleton)):
                 expr = expr.replace(expr[a1:a2 + 1], s2)
             binding[i] = expr
 
-        # Read the calendar type from the precipitation forcing NetCDF file
-        precipitation_map_path = binding["PrecipitationMaps"] + '.nc'
-        if not os.path.exists(precipitation_map_path):
-            raise LisfloodFileError(precipitation_map_path)
-        with Dataset(precipitation_map_path) as nc:
-            binding["calendar_type"] = get_calendar_type(nc)
+        # set the calendar type
+        binding["calendar_type"] = binding['CalendarConvention']
+
         return user, binding
 
     @staticmethod
@@ -399,9 +457,9 @@ class LisSettings(with_metaclass(Singleton)):
         res = {}
         repsteps = user_settings['ReportSteps'].split(',')
         if repsteps[0] == 'starttime':
-            repsteps[0] = bindings['StepStartInt']
+            repsteps[0] = str(bindings['StepStartInt'])
         if repsteps[-1] == 'endtime':
-            repsteps[-1] = bindings['StepEndInt']
+            repsteps[-1] = str(bindings['StepEndInt'])
         jjj = []
         for i in repsteps:
             if '..' in i:
@@ -462,6 +520,9 @@ class LisSettings(with_metaclass(Singleton)):
         # running through all times series
         timeseries = self.options['timeseries']
         for ts in timeseries.values():
+            ts_obj = self._set_active_options(ts, ts.repoption, ts.restrictoption)
+            if not ts_obj:
+                continue
             key = ts.name
             report_time_series_act[key] = self._set_active_options(ts, ts.repoption, ts.restrictoption)
 
@@ -535,7 +596,7 @@ def calendar(date_in, calendar_type='proleptic_gregorian'):
         # try reading a date in one of available formats
         try:
             _t_units = "hours since 1970-01-01 00:00:00"  # units used for date type conversion (datetime.datetime -> calendar-specific if needed)
-            date = parse_time_string(date_in, dayfirst=True)[0]  # datetime.datetime type
+            date = parsing.parse_time_string(date_in, dayfirst=True)[0]  # datetime.datetime type
             step = date2num(date, _t_units, calendar_type)  # float type
             return num2date(step, _t_units, calendar_type)  # calendar-dependent type from netCDF4.netcdftime._netcdftime module
         except:
@@ -569,7 +630,7 @@ def datetoint(date_in, binding=None):
     # DtDay = float(DtSec / 86400.)
     # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)
 
-    if type(date1) is datetime.datetime:
+    if isinstance(date1, (datetime.datetime, cftime.DatetimeProlepticGregorian, cftime.real_datetime)):
         str1 = date1.strftime("%d/%m/%Y %H:%M")
         # get total number of seconds corresponding to the time interval between dateIn and CalendarDayStart
         timeinterval_in_sec = int((date1 - begin).total_seconds())
@@ -631,7 +692,8 @@ class LisfloodRunInfo(Warning):
         nr_cores = settings.ncores[0]
         steps = len(settings.filter_steps)
         mode = self.modes[model.__class__.__name__]
-
+        activated_options = ','.join(k for k, v in option.items() if v and not k.startswith('rep') and 'netcdf' not in k.lower() and k not in ('timeseries',))
+        activated_rep = ','.join(k for k, v in option.items() if v and k.startswith('rep') and k not in 'timeseries')
         msg += "\t[X] LISFLOOD is used in the {}\n".format(mode)
         if option['InitLisflood']:
             msg += "\t[X] INITIALISATION RUN\n"
@@ -645,6 +707,8 @@ class LisfloodRunInfo(Warning):
         if mode in (self.modes['EnsKalmanFilterFramework'], self.modes['MonteCarloFramework']) and nr_cores > 1:
             msg += "\t[X] The simulation will try to use {} processors simultaneous\n".format(str(nr_cores))
         msg += "\t[X] The simulation output as specified in the settings file can be found in {}\n".format(out_dir)
+        msg += "\t[X] Activated modules: {}\n".format(activated_options)
+        msg += "\t[X] Report options: {}\n".format(activated_rep)
         self._msg = '{}{}'.format(header, msg)
 
     def __str__(self):
