@@ -58,7 +58,7 @@ class reservoir(HydroModule):
         settings = LisSettings.instance()
         option = settings.options
         maskinfo = MaskInfo.instance()
-        if option['simulateReservoirs']:
+        if option['simulateReservoirs'] and not option['InitLisflood']:
 
             # NoSubStepsRes=max(1,roundup(self.var.DtSec/loadmap('DtSecReservoirs')))
             # Number of sub-steps based on value of DtSecReservoirs,
@@ -70,18 +70,16 @@ class reservoir(HydroModule):
             self.var.ReservoirSitesC[self.var.ReservoirSitesC < 1] = 0
             self.var.ReservoirSitesC[self.var.IsChannel == 0] = 0
             # Get rid of any reservoirs that are not part of the channel network
-
-            # mask reserovoirs sites when using sub-catchments mask
-            # ReservoirSitesC_masked = ifthen(self.var.MaskMap,self.var.ReservoirSitesC)
             self.var.ReservoirSitesCC = np.compress(self.var.ReservoirSitesC > 0, self.var.ReservoirSitesC)
-            self.var.ReservoirIndex = np.nonzero(self.var.ReservoirSitesC)[0]
-            
-            if self.var.ReservoirSitesC.size == 0:
+            if self.var.ReservoirSitesCC.size == 0:
+                # break if no reservoirs
                 warnings.warn(LisfloodWarning('There are no reservoirs. Reservoirs simulation won\'t run'))
                 option['simulateReservoirs'] = False
                 option['repsimulateReservoirs'] = False
+                # rebuild lists of reported files with simulateReservoirs and repsimulateReservoirs = False
+                settings.build_reportedmaps_dicts()
                 return
-            # break if no reservoirs
+            self.var.ReservoirIndex = np.nonzero(self.var.ReservoirSitesC)[0]
 
             self.var.IsStructureKinematic = np.where(self.var.ReservoirSitesC > 0, np.bool8(1), self.var.IsStructureKinematic)
             # Add reservoir locations to structures map (used to modify LddKinematic
@@ -183,7 +181,7 @@ class reservoir(HydroModule):
         settings = LisSettings.instance()
         option = settings.options
         maskinfo = MaskInfo.instance()
-        if option['simulateReservoirs']:
+        if option['simulateReservoirs'] and not option['InitLisflood']:
             InvDtSecDay = 1 / float(86400)
             # InvDtSecDay=self.var.InvDtSec
             # ReservoirInflow = cover(ifthen(defined(self.var.ReservoirSites), upstream(
@@ -199,7 +197,13 @@ class reservoir(HydroModule):
 
             QResInM3Dt = ReservoirInflowCC * self.var.DtRouting
             # Reservoir inflow in [m3] per timestep (routing step)
-
+            
+            # print('RESERVOIRS MODULE IN')
+            # print(np.sum(self.var.ReservoirStorageM3))
+            if NoRoutingExecuted==0:
+                self.var.ReservoirStorageM3CC=np.compress(self.var.ReservoirSitesC > 0, self.var.ReservoirStorageM3)  ########################
+            # print(np.sum(self.var.ReservoirStorageM3CC))
+            
             self.var.ReservoirStorageM3CC += QResInM3Dt
             # New reservoir storage [m3] = plus inflow for this sub step
             self.var.ReservoirFillCC = self.var.ReservoirStorageM3CC / self.var.TotalReservoirStorageM3CC
@@ -262,14 +266,16 @@ class reservoir(HydroModule):
             # New reservoir fill
 
             # CM: Check ReservoirStorageM3CC for negative values and set them to zero
-            # FIXME
-            nel = len(self.var.ReservoirFillCC[:])  # always 1
-            for i in range(0, nel-1):  # never entering in loop
-                if np.isnan(self.var.ReservoirFillCC[i]) or self.var.ReservoirFillCC[i] < 0:
-                    msg = "Negative or NaN volume for reservoir fill set to 0. Increase computation time step for routing (DtSecChannel) \n"
-                    warnings.warn(LisfloodWarning(msg))
-                    self.var.ReservoirFillCC[self.var.ReservoirFillCC < 0] = 0
-                    self.var.ReservoirFillCC[np.isnan(self.var.ReservoirFillCC)] = 0
+            self.var.ReservoirFillCC[np.isnan(self.var.ReservoirFillCC)] = 0
+            self.var.ReservoirFillCC[self.var.ReservoirFillCC < 0] = 0
+
+            # nel = len(self.var.ReservoirFillCC[:])  # always 1 # FIXME
+            # for i in range(0, nel-1):  # never entering in loop
+            #     if np.isnan(self.var.ReservoirFillCC[i]) or self.var.ReservoirFillCC[i] < 0:
+            #         msg = "Negative or NaN volume for reservoir fill set to 0. Increase computation time step for routing (DtSecChannel) \n"
+            #         warnings.warn(LisfloodWarning(msg))
+            #         self.var.ReservoirFillCC[self.var.ReservoirFillCC < 0] = 0
+            #         self.var.ReservoirFillCC[np.isnan(self.var.ReservoirFillCC)] = 0
 
             # Check ReservoirFillCC for negative values and set them to zero
             # if np.isnan(self.var.ReservoirFillCC).any() or (self.var.ReservoirFillCC < 0).any():
@@ -296,6 +302,8 @@ class reservoir(HydroModule):
                     self.var.sumResOutCC += QResOutM3DtCC
                     # summing up over all sub timesteps
 
+
+
             if NoRoutingExecuted == (self.var.NoRoutSteps-1):
 
                 # expanding the size after last sub timestep
@@ -304,6 +312,11 @@ class reservoir(HydroModule):
                 np.put(self.var.ReservoirStorageM3, self.var.ReservoirIndex, self.var.ReservoirStorageM3CC)
                 np.put(self.var.ReservoirFill, self.var.ReservoirIndex, self.var.ReservoirFillCC)
 
+                # print('RESERVOIRS MODULE OUT')
+                # print(np.sum(self.var.ReservoirStorageM3))
+                # print(np.sum(self.var.ReservoirStorageM3CC))
+
+
                 if option['repsimulateReservoirs']:
                     np.put(self.var.ReservoirInflowM3S, self.var.ReservoirIndex, self.var.sumResInCC / self.var.DtSec)
                     np.put(self.var.ReservoirOutflowM3S, self.var.ReservoirIndex, self.var.sumResOutCC / self.var.DtSec)
