@@ -22,6 +22,17 @@ import numpy as np
 from ..global_modules.settings import MaskInfo, LisSettings
 from ..global_modules.add1 import loadmap, defsoil
 from . import HydroModule
+from collections import OrderedDict
+#from global_modules.add1 import *
+import xarray as xr
+
+
+def pressure2SoilMoistureFun(residual_sm, sat_sm, GenuA, GenuN, GenuM):
+    """Generate a function to compute soil moisture values corresponding to characteristic pressure head levels [cm].
+       IT WON'T BE NEEDED ANYMORE ONCE DATA STRUCTURES WILL ALSO HAVE A "layer" DIMENSION, TO MERGE LAYER-SPECIFIC VARIABLES (1a, 1b, 2)"""
+    def Mualem(pressure):
+        return residual_sm + (sat_sm - residual_sm) / ((1 + (GenuA * pressure) ** GenuN) ** GenuM)
+    return Mualem
 
 
 class soil(HydroModule):
@@ -60,8 +71,6 @@ class soil(HydroModule):
     def initial(self):
         """ initial part of the soil module
         """
-        maskinfo = MaskInfo.instance()
-
         def splitlanduse(array1, array2=None, array3=None):
             """ splits maps into the 3 different land use types - other , forest, irrigation
             """
@@ -76,8 +85,13 @@ class soil(HydroModule):
         # using maps instead of table to be more sensitive to landuse change
         # Soil properties for percentage of pixel area without forest, water ,
         # impervious soil
+        settings = LisSettings.instance()
+        option = settings.options
+        binding = settings.binding
+        maskinfo = MaskInfo.instance()
+        if not option["cropsEPIC"]: # If EPIC is active, the rice fraction initialisation is handled by EPIC (setSoilFractions in EPIC_main.py)
+            self.var.SoilFraction.values[self.var.vegetation.index('Rainfed_prescribed')] += self.var.RiceFraction
 
-        self.var.OtherFraction += self.var.RiceFraction
         # for the moment rice is treated as other fraction
         # if the fraction of water varies then the other fraction are stored
         self.var.WaterFractionBase = self.var.WaterFraction.copy()
@@ -85,8 +99,6 @@ class soil(HydroModule):
         self.var.IrrigationFractionBase = self.var.IrrigationFraction.copy()
         self.var.ForestFractionBase = self.var.ForestFraction.copy()
         self.var.DirectRunoffFractionBase = self.var.DirectRunoffFraction.copy()
-
-        self.var.SoilFraction = self.var.ForestFraction + self.var.OtherFraction + self.var.IrrigationFraction
         self.var.PermeableFraction = 1 - self.var.DirectRunoffFraction - self.var.WaterFraction
         # Permeable fraction of pixel
 
@@ -94,10 +106,10 @@ class soil(HydroModule):
         # soil depth top layer for every landuse but forest, impervious soil
         # and water
 
-        self.var.SoilDepth1a = defsoil('SoilDepth1', 'SoilDepth1Forest')
-        self.var.SoilDepth1b = defsoil('SoilDepth2', 'SoilDepth2Forest')
-        self.var.SoilDepth2 = defsoil('SoilDepth3', 'SoilDepth3Forest')
-
+        self.var.SoilDepth1a = self.var.defsoil('SoilDepth1', 'SoilDepth1Forest')
+        self.var.SoilDepth1b = self.var.defsoil('SoilDepth2', 'SoilDepth2Forest')
+        self.var.SoilDepth2 = self.var.defsoil('SoilDepth3', 'SoilDepth3Forest')
+        self.var.SoilDepthTotal = self.var.SoilDepth1a + self.var.SoilDepth1b + self.var.SoilDepth2
         # ----------------- miscParameters ---------------------------
 
         self.var.CourantCrit = loadmap('CourantCrit')
@@ -119,14 +131,13 @@ class soil(HydroModule):
         # using maps instead of table to be more sensitive to landuse change
         # Soil properties for percentage of pixel area without forest, water ,
         # impervious soil
-        self.var.CropCoef = defsoil('MapCropCoef', 'MapForestCropCoef', 'MapIrrigationCropCoef')
+        self.var.CropCoef = self.var.defsoil('MapCropCoef', 'MapForestCropCoef', 'MapIrrigationCropCoef')
 
         # crop coefficients for each land use class
-        self.var.CropGroupNumber = defsoil('MapCropGroupNumber', 'MapForestCropGroupNumber', 'MapIrrigationCropGroupNumber')
+        self.var.CropGroupNumber = self.var.defsoil('MapCropGroupNumber', 'MapForestCropGroupNumber', 'MapIrrigationCropGroupNumber')
 
-        # crop group numbers for soil water depletion factor
-        # self.var.NManning = [loadmap('MapN'), loadmap('MapForestN'), 0.02]
-        self.var.NManning = defsoil('MapN', 'MapForestN', 0.02)
+        # crop group numbers for soil water depletion factor        
+        self.var.NManning = self.var.defsoil('MapN','MapForestN', 0.02, OrderedDict([self.var.dim_runoff, self.var.dim_pixel]))
         # Manning's roughness coefficient for each land use class
         # third manning is from direct runoff
         # self.var.NManningDirect=scalar(0.02)
@@ -137,92 +148,140 @@ class soil(HydroModule):
 
         # HYPRES Parameters for TopSoil:
 
-        self.var.KSat1a = defsoil('MapKSat1', 'MapKSat1Forest')
-        self.var.KSat1b = defsoil('MapKSat2', 'MapKSat2Forest')
-        self.var.KSat2 = defsoil('MapKSat3')
+        self.var.KSat1a = self.var.defsoil('MapKSat1', 'MapKSat1Forest')
+        self.var.KSat1b = self.var.defsoil('MapKSat2', 'MapKSat2Forest')
+        self.var.KSat2 = self.var.defsoil('MapKSat3')
         # Saturated conductivity for other land use and forest
-        Lambda1a = defsoil('MapLambda1', 'MapLambda1Forest')
-        Lambda1b = defsoil('MapLambda2', 'MapLambda2Forest')
-        Lambda2 = defsoil('MapLambda3')
+        Lambda1a = self.var.defsoil('MapLambda1', 'MapLambda1Forest')
+        Lambda1b = self.var.defsoil('MapLambda2', 'MapLambda2Forest')
+        Lambda2 = self.var.defsoil('MapLambda3')
         # Pore-size index (for Van Genuchten parameters)
-        GenuAlpha1a = defsoil('MapGenuAlpha1', 'MapGenuAlpha1Forest')
-        GenuAlpha1b = defsoil('MapGenuAlpha2', 'MapGenuAlpha2Forest')
-        GenuAlpha2 = defsoil('MapGenuAlpha3')
+        GenuAlpha1a = self.var.defsoil('MapGenuAlpha1', 'MapGenuAlpha1Forest')
+        GenuAlpha1b = self.var.defsoil('MapGenuAlpha2', 'MapGenuAlpha2Forest')
+        GenuAlpha2 = self.var.defsoil('MapGenuAlpha3')
         # Van Genuchten Alpha coefficient
-        ThetaS1a = defsoil('MapThetaSat1','MapThetaSat1Forest')
-        ThetaS1b = defsoil('MapThetaSat2','MapThetaSat2Forest')
-        ThetaS2 = defsoil('MapThetaSat3')
+        ThetaS1a = self.var.defsoil('MapThetaSat1','MapThetaSat1Forest')
+        ThetaS1b = self.var.defsoil('MapThetaSat2','MapThetaSat2Forest')
+        ThetaS2 = self.var.defsoil('MapThetaSat3')
         # Soil moisture content at saturation [V/V]
-        ThetaRes1a = defsoil('MapThetaRes1', 'MapThetaRes1Forest')
-        ThetaRes1b = defsoil('MapThetaRes2', 'MapThetaRes2Forest')
-        ThetaRes2 = defsoil('MapThetaRes3')
+        ThetaRes1a = self.var.defsoil('MapThetaRes1', 'MapThetaRes1Forest')
+        ThetaRes1b = self.var.defsoil('MapThetaRes2', 'MapThetaRes2Forest')
+        ThetaRes2 = self.var.defsoil('MapThetaRes3')
         # Soil moisture content at saturation [V/V]
 
-        # GenuN1=Lambda+1
-        GenuN1a = [x + 1 for x in Lambda1a]
-        GenuN1b = [x + 1 for x in Lambda1b]
-        GenuN2 = [x + 1 for x in Lambda2]
+        GenuN1a = 1 + Lambda1a
+        GenuN1b = 1 + Lambda1b
+        GenuN2 = 1 + Lambda2
         # self.var.GenuM1=Lambda1/GenuN1
-        self.var.GenuM1a = [x / y for x, y in zip(Lambda1a, GenuN1a)]
-        self.var.GenuM1b = [x / y for x, y in zip(Lambda1b, GenuN1b)]
-        self.var.GenuM2 = [x / y for x, y in zip(Lambda2, GenuN2)]
+        self.var.GenuM1a = Lambda1a / GenuN1a
+        self.var.GenuM1b = Lambda1b / GenuN1b
+        self.var.GenuM2 = Lambda2 / GenuN2
         # self.var.GenuInvM1=1/self.var.GenuM1
-        self.var.GenuInvM1a = [1 / x for x in self.var.GenuM1a]
-        self.var.GenuInvM1b = [1 / x for x in self.var.GenuM1b]
-        self.var.GenuInvM2 = [1 / x for x in self.var.GenuM2]
+        self.var.GenuInvM1a = 1 / self.var.GenuM1a
+        self.var.GenuInvM1b = 1 / self.var.GenuM1b
+        self.var.GenuInvM2 = 1 / self.var.GenuM2
         # self.var.GenuInvN1=1/GenuN1
-        self.var.GenuInvN1a = [1 / x for x in GenuN1a]
-        self.var.GenuInvN1b = [1 / x for x in GenuN1b]
-        self.var.GenuInvN2 = [1 / x for x in GenuN2]
+        self.var.GenuInvN1a = 1 / GenuN1a
+        self.var.GenuInvN1b = 1 / GenuN1b
+        self.var.GenuInvN2 = 1 / GenuN2
         # GenuInvAlpha1=1/GenuAlpha1
-        self.var.GenuInvAlpha1a = [1 / x for x in GenuAlpha1a]
-        self.var.GenuInvAlpha1b = [1 / x for x in GenuAlpha1b]
-        self.var.GenuInvAlpha2 = [1 / x for x in GenuAlpha2]
-        # Van Genuchten n and m coefficients
-        # In previous versions both n and m were read from separate tables.
-        # Since both are calculated from same pore-size index (Lambda=n-1),
-        # only table of Lambda is used here (Lambda is also more commonly
-        # reported in the literature, which makes any comparison easier)
-        # Inverse values computed once here for improved performance
-        # (inverse of N and Alpha only needed if option simulatePF is used)
-        # self.var.WS1=ThetaS1*self.var.SoilDepth1
-        self.var.WS1a = [x * y for x, y in zip(ThetaS1a, self.var.SoilDepth1a)]
-        self.var.WS1b = [x * y for x, y in zip(ThetaS1b, self.var.SoilDepth1b)]
-        self.var.WS2 = [x * y for x, y in zip(ThetaS2, self.var.SoilDepth2)]
-        self.var.WS1 = np.add(self.var.WS1a, self.var.WS1b)
-        # self.var.WRes1=ThetaRes1*self.var.SoilDepth1
-        self.var.WRes1a = [x * y for x, y in zip(ThetaRes1a, self.var.SoilDepth1a)]
-        self.var.WRes1b = [x * y for x, y in zip(ThetaRes1b, self.var.SoilDepth1b)]
-        self.var.WRes1 = np.add(self.var.WRes1a, self.var.WRes1b)
-        self.var.WRes2 = [x * y for x, y in zip(ThetaRes2, self.var.SoilDepth2)]
+        self.var.GenuInvAlpha1a = 1 / GenuAlpha1a
+        self.var.GenuInvAlpha1b = 1 / GenuAlpha1b
+        self.var.GenuInvAlpha2 = 1 / GenuAlpha2
+        
+        # Soil moisture [mm] at saturation:
+        self.var.WS1a = ThetaS1a * self.var.SoilDepth1a
+        self.var.WS1b = ThetaS1b * self.var.SoilDepth1b
+        self.var.WS2 = ThetaS2 * self.var.SoilDepth2
+        self.var.WS1 = self.var.WS1a + self.var.WS1b
+        # Residual soil moisture [mm]:
+        self.var.WRes1a = ThetaRes1a * self.var.SoilDepth1a
+        self.var.WRes1b = ThetaRes1b * self.var.SoilDepth1b
+        self.var.WRes2 = ThetaRes2 * self.var.SoilDepth2
+        self.var.WRes1 = self.var.WRes1a + self.var.WRes1b
+        # Saturation minus residual soil moisture [mm]:
+        self.var.WS1WResa = self.var.WS1a - self.var.WRes1a
+        self.var.WS1WResb = self.var.WS1b - self.var.WRes1b
+        self.var.WS2WRes = self.var.WS2 - self.var.WRes2
+        
+        # Generate Mualem functions (van Genuchten, 1980) for the 3 soil layers:
+        MualemEq1a = pressure2SoilMoistureFun(self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a)
+        MualemEq1b = pressure2SoilMoistureFun(self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b)
+        MualemEq2 = pressure2SoilMoistureFun(self.var.WRes2, self.var.WS2, GenuAlpha2, GenuN2, self.var.GenuM2)
+        # Soil moisture at field capacity (pF2, 100 cm) [mm water slice]:       
+        self.var.WFC1a = MualemEq1a(100)
+        self.var.WFC1b = MualemEq1b(100)
+        self.var.WFC2 = MualemEq2(100)
+        self.var.WFC1 = self.var.WFC1a + self.var.WFC1b
+        # Soil moisture [mm] at pF3, 1000 cm (pressure):
+        self.var.WPF3a = MualemEq1a(1000)
+        self.var.WPF3b = MualemEq1b(1000)
+        self.var.WPF3 = self.var.WPF3a + self.var.WPF3b
+        # Soil moisture at wilting point (pF4.2, 10**4.2 cm) in [mm] water slice:
+        self.var.WWP1a = MualemEq1a(15000)
+        self.var.WWP1b = MualemEq1b(15000)
+        self.var.WWP2 = MualemEq2(15000)
+        self.var.WWP1 = self.var.WWP1a + self.var.WWP1b       
+       
+        self.var.PoreSpaceNotZero1a = np.logical_and(self.var.SoilDepth1a != 0, self.var.WS1a != 0)
+        self.var.PoreSpaceNotZero1b = np.logical_and(self.var.SoilDepth1b != 0, self.var.WS1b != 0)
+        self.var.PoreSpaceNotZero2 = np.logical_and(self.var.SoilDepth2 != 0, self.var.WS2 != 0)
 
-        # self.var.WS1WRes=self.var.WS1-self.var.WRes1
-        self.var.WS1WResa = [x - y for x, y in zip(self.var.WS1a, self.var.WRes1a)]
-        self.var.WS1WResb = [x - y for x, y in zip(self.var.WS1b, self.var.WRes1b)]
-        self.var.WS2WRes = [x - y for x, y in zip(self.var.WS2, self.var.WRes2)]
+        # ***** INITIAL VALUES for soil
+        
+        if not option["cropsEPIC"]: 
+       
+           ThetaInit1aValue = self.var.allocateVariableAllVegetation()
+           ThetaInit1bValue = self.var.allocateVariableAllVegetation()
+           ThetaInit2Value = self.var.allocateVariableAllVegetation()
 
-        # Express volume fractions in [mm] water slice
-        # self.var.WFC1=self.var.WRes1+(self.var.WS1-self.var.WRes1)/((1+(GenuAlpha1*100)**GenuN1)**self.var.GenuM1)
-        self.var.WFC1a = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 100) ** gn) ** gm), self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a))
-        self.var.WFC1b = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 100) ** gn) ** gm), self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b))
-        self.var.WFC1 = np.add(self.var.WFC1a, self.var.WFC1b)
-        self.var.WFC2 = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 100) ** gn) ** gm), self.var.WRes2, self.var.WS2, GenuAlpha2, GenuN2, self.var.GenuM2))
-        # Soil moisture at field capacity (pF2, 100 cm) [mm water slice]
-        # Mualem equation (van Genuchten, 1980)
+           iveg = self.var.vegetation.index('Rainfed_prescribed')           
+           ThetaInit1aValue.values[iveg] = loadmap('ThetaInit1Value')
+           ThetaInit1bValue.values[iveg] = loadmap('ThetaInit2Value')
+           ThetaInit2Value.values[iveg] = loadmap('ThetaInit3Value')
+           
+           iveg = self.var.vegetation.index('Forest_prescribed')
+           ThetaInit1aValue.values[iveg] = loadmap('ThetaForestInit1Value')
+           ThetaInit1bValue.values[iveg] = loadmap('ThetaForestInit2Value')
+           ThetaInit2Value.values[iveg] = loadmap('ThetaForestInit3Value')
+           
+           iveg = self.var.vegetation.index('Irrigated_prescribed')
+           ThetaInit1aValue.values[iveg] = loadmap('ThetaIrrigationInit1Value')
+           ThetaInit1bValue.values[iveg] = loadmap('ThetaIrrigationInit2Value')
+           ThetaInit2Value.values[iveg] = loadmap('ThetaIrrigationInit3Value')
+        else:
+            ThetaInit1aValue = self.var.initialiseVariableAllVegetation('ThetaInit1Value')
+            ThetaInit1bValue = self.var.initialiseVariableAllVegetation('ThetaInit2Value')
+            ThetaInit2Value = self.var.initialiseVariableAllVegetation('ThetaInit3Value')
+        
+  
+        # Allocate soil moisture [mm] data structures (merge them into a single xarray.DataArray by adding a "layer" dimension !!!)
+        self.var.W1a = self.var.allocateVariableAllVegetation()
+        self.var.W1b = self.var.allocateVariableAllVegetation()
+        self.var.W2 = self.var.allocateVariableAllVegetation()
+        # Initialise soil moisture in permeable fraction of pixel:
+        # Set soil to field capacity if initial value in settings file equals -9999;
+        # Set to zero if soil depth is zero.
+        # IMPORTANT: WInit1 and WInit2 represent the soil moisture in the *permeable* fraction of the pixel *only*
+        # (since all soil moisture-related calculations are done for permeable fraction only!).
 
-        self.var.WPF3a = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 1000) ** gn) ** gm), self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a))
-        self.var.WPF3b = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 1000) ** gn) ** gm), self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b))
-        self.var.WPF3 = np.add(self.var.WPF3a, self.var.WPF3b)
-
-        # Soil moisture at field capacity (pF3, 1000 cm) [mm water slice]
-        # Mualem equation (van Genuchten, 1980)        # self.var.WWP1=self.var.WRes1+(self.var.WS1-self.var.WRes1)/  ((1+(GenuAlpha1*(10**4.2))**GenuN1)**self.var.GenuM1)
-        self.var.WWP1a = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 15000) ** gn) ** gm), self.var.WRes1a, self.var.WS1a, GenuAlpha1a, GenuN1a, self.var.GenuM1a))
-        self.var.WWP1b = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 15000) ** gn) ** gm), self.var.WRes1b, self.var.WS1b, GenuAlpha1b, GenuN1b, self.var.GenuM1b))
-        self.var.WWP1 = np.add(self.var.WWP1a, self.var.WWP1b)
-        self.var.WWP2 = list(map(lambda res, s, ga, gn, gm: res + (s - res) / ((1 + (ga * 15000) ** gn) ** gm), self.var.WRes2, self.var.WS2, GenuAlpha2, GenuN2, self.var.GenuM2))
-        # Soil moisture at wilting point (pF4.2, 10**4.2 cm) in [mm] water slice
-        # Mualem equation (van Genuchten, 1980)
+        for veg, luse in self.var.VEGETATION_LANDUSE.items():
+            iveg = self.var.vegetation.index(veg)
+            iluse = self.var.SOIL_USES.index(luse)
+            ini_1a = np.where(ThetaInit1aValue[iveg] == -9999, self.var.WFC1a[iluse], ThetaInit1aValue[iveg] * self.var.SoilDepth1a[iluse])
+            self.var.W1a[iveg] = np.where(self.var.PoreSpaceNotZero1a[iluse], ini_1a, 0)
+            ini_1b = np.where(ThetaInit1bValue[iveg] == -9999, self.var.WFC1b[iluse], ThetaInit1bValue[iveg] * self.var.SoilDepth1b[iluse])
+            self.var.W1b[iveg] = np.where(self.var.PoreSpaceNotZero1b[iluse], ini_1b, 0)
+            ini_2 = np.where(ThetaInit2Value[iveg] == -9999, self.var.WFC2[iluse], ThetaInit2Value[iveg] * self.var.SoilDepth2[iluse])
+            self.var.W2[iveg] = np.where(self.var.PoreSpaceNotZero2[iluse], ini_2, 0)
+        self.var.W1 = self.var.W1a + self.var.W1b
+                    
 
+        self.var.Sat1a = self.var.allocateVariableAllVegetation()
+        self.var.Sat1b = self.var.allocateVariableAllVegetation()
+        self.var.Sat1 = self.var.allocateVariableAllVegetation()
+        self.var.Sat2 = self.var.allocateVariableAllVegetation()
+        '''
         self.var.PoreSpaceNotZero1a = splitlanduse(np.bool8((self.var.SoilDepth1a[0] != 0) & (self.var.WS1a[0] != 0)),
                                                    np.bool8((self.var.SoilDepth1a[1] != 0) & (self.var.WS1a[1] != 0)),
                                                    np.bool8((self.var.SoilDepth1a[2] != 0) & (self.var.WS1a[2] != 0)))
@@ -285,19 +344,23 @@ class soil(HydroModule):
         self.var.Sat1b = splitlanduse(maskinfo.in_zero())
         self.var.Sat1 = splitlanduse(maskinfo.in_zero())
         self.var.Sat2 = splitlanduse(maskinfo.in_zero())
-
+        '''
 
         # ************************************************************
         # ***** REPEATEDLY USED EXPRESSIONS IN XINANJIANG
         # ***** INFILTRATION MODEL
         # ************************************************************
         self.var.b_Xinanjiang = loadmap('b_Xinanjiang')
+    
+        if isinstance(self.var.b_Xinanjiang, float):
+           MAP = []
+           MAP = maskinfo.in_zero() +   self.var.b_Xinanjiang
+           self.var.b_Xinanjiang = []
+           self.var.b_Xinanjiang = MAP
+               
         self.var.PowerInfPot = (self.var.b_Xinanjiang + 1) / self.var.b_Xinanjiang
         # Power in infiltration equation
-        # self.var.StoreMaxPervious=self.var.WS1/(self.var.b_Xinanjiang+1)
-#        self.var.StoreMaxPervious = [x / (self.var.b_Xinanjiang + 1) for x in self.var.WS1a]
-
-        self.var.StoreMaxPervious = [x / (self.var.b_Xinanjiang + 1) for x in self.var.WS1]
+        self.var.StoreMaxPervious = self.var.WS1 / (self.var.b_Xinanjiang + 1)
         # Maximum soil moisture storage in pervious fraction of
         # pixel (expressed as [mm] water slice)
 
@@ -305,63 +368,82 @@ class soil(HydroModule):
         # ****  PowerPrefFlow
         # ********************************************
         self.var.PowerPrefFlow = loadmap('PowerPrefFlow')
+                
+        if isinstance(self.var.PowerPrefFlow, float):
+           MAP = []
+           MAP = maskinfo.in_zero() +   self.var.PowerPrefFlow
+           self.var.PowerPrefFlow = []
+           self.var.PowerPrefFlow = MAP
 
+           
         # ************************************************************
         # ***** INITIAL VALUES
         # ************************************************************
         # Inputs in waterbalance model and/or initial assumption
         # CMmod
         # DSLRInit = defsoil('DSLRInitValue', 'DSLRForestInitValue','DSLRIrrigationInitValue')
+        
+        
+        if not option["cropsEPIC"]: 
+           self.var.DSLR = self.var.allocateVariableAllVegetation()
+           ivegRainfedPrescribed = self.var.vegetation.index('Rainfed_prescribed')
+           ivegForestPrescribed = self.var.vegetation.index('Forest_prescribed')
+           ivegIrrigatedPrescribed = self.var.vegetation.index('Irrigated_prescribed')
 
-        DSLRInit = defsoil('DSLRInitValue', 'DSLRForestInitValue', 'DSLRIrrigationInitValue')
-        self.var.DSLR = splitlanduse(*[np.maximum(d_i, 1) for d_i in DSLRInit])
-        # Days since last rainfall
-        # Impose lower limit of 1 because otherwise
-        # square root may be taken out of a negative number for Dtsec lt 86400!
+           self.var.DSLR[ivegRainfedPrescribed] = loadmap('DSLRInitValue')
+           self.var.DSLR[ivegForestPrescribed]  = loadmap('DSLRForestInitValue')
+           self.var.DSLR[ivegIrrigatedPrescribed]  = loadmap('DSLRIrrigationInitValue')
+           self.var.DSLR[ivegRainfedPrescribed]  = np.where(self.var.DSLR[ivegRainfedPrescribed]<1,1,self.var.DSLR[ivegRainfedPrescribed])
+           self.var.DSLR[ivegForestPrescribed]  = np.where(self.var.DSLR[ivegForestPrescribed]<1,1,self.var.DSLR[ivegForestPrescribed])
+           self.var.DSLR[ivegIrrigatedPrescribed]  = np.where(self.var.DSLR[ivegIrrigatedPrescribed]<1,1,self.var.DSLR[ivegIrrigatedPrescribed])
+        else:
+           self.var.DSLR = np.maximum(self.var.initialiseVariableAllVegetation('DSLRInitValue'), 1) # DSLR<1 may cause square root of negative number, if Dtsec > 86400!
+        
+        if not option["cropsEPIC"]:  
+           self.var.CumInterception = self.var.allocateVariableAllVegetation()
+           self.var.CumInterception[ivegRainfedPrescribed] = loadmap('CumIntInitValue')
+           self.var.CumInterception[ivegForestPrescribed] = loadmap('CumIntForestInitValue')
+           self.var.CumInterception[ivegIrrigatedPrescribed] = loadmap('CumIntIrrigationInitValue')
+        else:
+           self.var.CumInterception = self.var.initialiseVariableAllVegetation('CumIntInitValue')
+        
+        # Initialising cumulative output variables  needed to compute the cumulative mass balance error
+        self.var.TotalPrecipitation =  maskinfo.in_zero() # precipitation [mm]
+        self.var.TaCUM =  maskinfo.in_zero() # Cumulative transpiration [mm]
+        self.var.TaWB = maskinfo.in_zero() # Cumulative transpiration [mm] at the end of the computational time step (waterbalance.py)
+        self.var.TaInterceptionCUM =  maskinfo.in_zero() # Cumulative evaporation from interception store [mm]
+        self.var.TaInterceptionWB =  maskinfo.in_zero() # Evaporation from interception store [mm] at the end of the computational time step (waterbalance.py)
+        self.var.ESActCUM =  maskinfo.in_zero() # Cumulative evaporation [mm]
+        self.var.ESActWB = maskinfo.in_zero() # Cumulative evaporation [mm] at the end of the computational time step (waterbalance.py)
+        self.var.SoilMoistureStressDays = self.var.allocateVariableAllVegetation() # number of days in simulation with soil moisture stress (days)
+        self.var.Theta1a = self.var.allocateVariableAllVegetation() # Theta values are just allocated here - their values are dynamically compute in soilloop.py
+        self.var.Theta1b = self.var.allocateVariableAllVegetation()
+        self.var.Theta2 = self.var.allocateVariableAllVegetation()
+        self.var.TaInterception = self.var.allocateVariableAllVegetation()
+        self.var.Interception = self.var.allocateVariableAllVegetation()
+        self.var.LeafDrainage = self.var.allocateVariableAllVegetation()
+        self.var.potential_transpiration = self.var.allocateVariableAllVegetation()
+        self.var.Ta = self.var.allocateVariableAllVegetation()
+        self.var.ESAct = self.var.allocateVariableAllVegetation()
+        self.var.PrefFlow = self.var.allocateVariableAllVegetation()
+        self.var.Infiltration = self.var.allocateVariableAllVegetation()
+        self.var.SeepTopToSubA = self.var.allocateVariableAllVegetation()
+        self.var.SeepTopToSubB = self.var.allocateVariableAllVegetation()
+        self.var.SeepSubToGW = self.var.allocateVariableAllVegetation()
+        self.var.Theta = self.var.allocateVariableAllVegetation()
+        self.var.AvailableWaterForInfiltration = self.var.allocateVariableAllVegetation()
+        ivegvalues = []
+        for veg in self.var.prescribed_vegetation:
+            ivegvalues.append(self.var.vegetation.index(veg))
+        self.var.RWS = self.var.allocateVariableAllVegetation()[ivegvalues]
 
-        # CumInterceptionInit=ifthen(defined(MaskMap),scalar(loadmap('CumIntInitValue')))
-        self.var.CumInterception = defsoil('CumIntInitValue', 'CumIntForestInitValue', 'CumIntIrrigationInitValue')
-        # initial cumulative interception, [mm]
-
-        # Initialising cumulative output variables
-        # These are all needed to compute the cumulative mass balance error
-
-        self.var.TotalPrecipitation = maskinfo.in_zero()
-        # precipitation [mm]
-        self.var.TaCUM = maskinfo.in_zero()
-        # Cumulative transpiration [mm]
-        self.var.TaInterceptionCUM = maskinfo.in_zero()
-        # Cumulative evaporation from interception store [mm]
-        self.var.ESActCUM = maskinfo.in_zero()
-        # Cumulative evaporation [mm]
-
-        self.var.SoilMoistureStressDays = splitlanduse(maskinfo.in_zero())
-        # number of days in simulation with soil moisture stress (days)
-
-        self.var.Theta1a = splitlanduse(maskinfo.in_zero())
-        self.var.Theta1b = splitlanduse(maskinfo.in_zero())
-        self.var.Theta2 = splitlanduse(maskinfo.in_zero())
-
-        self.var.TaInterception = defsoil(maskinfo.in_zero())
-        self.var.Interception = defsoil(maskinfo.in_zero())
-        self.var.LeafDrainage = defsoil(maskinfo.in_zero())
-        self.var.Ta = splitlanduse(maskinfo.in_zero())
-        self.var.ESAct = splitlanduse(maskinfo.in_zero())
-        self.var.PrefFlow = splitlanduse(maskinfo.in_zero())
-        self.var.Infiltration = splitlanduse(maskinfo.in_zero())
-        self.var.SeepTopToSubA = splitlanduse(maskinfo.in_zero())
-        self.var.SeepTopToSubB = splitlanduse(maskinfo.in_zero())
-        self.var.SeepSubToGW = splitlanduse(maskinfo.in_zero())
-
-        self.var.Theta = splitlanduse(maskinfo.in_zero())
-        self.var.AvailableWaterForInfiltration = splitlanduse(maskinfo.in_zero())
-        self.var.RWS = splitlanduse(maskinfo.in_zero())
 
         # ************************************************************
         # ***** INITIALIZATION OF IMPERVIOUS SOIL     ****************
         # ************************************************************
 
         self.var.CumInterSealed = loadmap('CumIntSealedInitValue')
+        
         # cumulative depression storage
         self.var.SMaxSealed = loadmap('SMaxSealed')
         # maximum depression storage for water on impervious surface
@@ -372,18 +454,19 @@ class soil(HydroModule):
 
         if option['drainedIrrigation']:
             self.var.DrainedFraction = loadmap('DrainedFraction')
+        else: 
+            self.var.DrainedFraction = 0.0 ##### STEF: the master simply do not assign this!
             # drained fraction: water from irrigated water is directly send to channel
             # depending on the drained fraction
 
 # ************************************************************
 # ***** INITIALIZATION OF PF calculation      ****************
 # ************************************************************
-
         if option['simulatePF']:
             self.var.HeadMax = loadmap('HeadMax')
-            self.var.pF0 = defsoil(maskinfo.in_zero())
-            self.var.pF1 = defsoil(maskinfo.in_zero())
-            self.var.pF2 = defsoil(maskinfo.in_zero())
+            self.var.pF0 = self.var.allocateVariableAllVegetation()
+            self.var.pF1 = self.var.allocateVariableAllVegetation()
+            self.var.pF2 = self.var.allocateVariableAllVegetation()
 
     def dynamic_perpixel(self):
         """ dynamic part of the soil module
@@ -391,16 +474,19 @@ class soil(HydroModule):
         """
         self.var.TaInterceptionAll = self.var.deffraction(self.var.TaInterception) + self.var.DirectRunoffFraction * self.var.TASealed
         self.var.TaInterceptionCUM += self.var.TaInterceptionAll
+        self.var.TaInterceptionWB = self.var.TaInterceptionAll
         # Cumulative evaporation of intercepted water [mm]
 
         self.var.TaPixel = self.var.deffraction(self.var.Ta)
         # pixel-average transpiration in [mm] per timestep
         # (no transpiration from direct runoff fraction)
         self.var.TaCUM += self.var.TaPixel
+        self.var.TaWB = self.var.TaPixel
         self.var.ESActPixel = self.var.deffraction(self.var.ESAct) + self.var.WaterFraction * self.var.EWaterAct
         # Pixel-average soil evaporation in [mm] per time step
         # (no evaporation from direct runoff fraction)
         self.var.ESActCUM += self.var.ESActPixel
+        self.var.ESActWB = self.var.ESActPixel
 
         self.var.PrefFlowPixel = self.var.deffraction(self.var.PrefFlow)
         # Pixel-average preferential flow in [mm] per timestep
@@ -408,14 +494,21 @@ class soil(HydroModule):
         self.var.InfiltrationPixel = self.var.deffraction(self.var.Infiltration)
         # Pixel-average infiltration in [mm] per timestep
         # (no infiltration in direct runoff fraction)
-
-        self.var.Theta[0] = self.var.OtherFraction * (self.var.W1a[0] + self.var.W1b[0] + self.var.W2[0]) / (self.var.SoilDepth1a[0] + self.var.SoilDepth1b[0] + self.var.SoilDepth2[0])
-        self.var.Theta[1] = self.var.ForestFraction * (self.var.W1a[1] + self.var.W1b[1] + self.var.W2[1]) / (self.var.SoilDepth1a[1] + self.var.SoilDepth1b[1] + self.var.SoilDepth2[1])
-        self.var.Theta[2] = self.var.IrrigationFraction * (self.var.W1a[2] + self.var.W1b[2] + self.var.W2[2]) / (self.var.SoilDepth1a[2] + self.var.SoilDepth1b[2] + self.var.SoilDepth2[2])
-        self.var.ThetaAll = np.where(self.var.SoilFraction > 0, (self.var.Theta[0] + self.var.Theta[1] + self.var.Theta[2]) / self.var.SoilFraction, 0.0)
+        tot_sm = self.var.W1a + self.var.W1b + self.var.W2
+        for landuse, veg_list in self.var.LANDUSE_VEGETATION.items():
+            iveg_list,iveg_list_pres,ilanduse = self.var.get_indexes_from_landuse_and_veg_list_GLOBAL(landuse, veg_list)
+            self.var.Theta[iveg_list] = self.var.SoilFraction[iveg_list_pres] * tot_sm[iveg_list] / self.var.SoilDepthTotal[ilanduse]
+        ax_veg=self.var.SoilFraction.dims.index("vegetation")
+        soil_fract_sum = np.sum(self.var.SoilFraction,ax_veg)
+        self.var.ThetaAll = np.where(soil_fract_sum > 0, np.sum(self.var.Theta,ax_veg) / soil_fract_sum, 0)
 
         self.var.SeepTopToSubPixelA = self.var.deffraction(self.var.SeepTopToSubA)
         self.var.SeepTopToSubPixelB = self.var.deffraction(self.var.SeepTopToSubB)
         self.var.SeepSubToGWPixel = self.var.deffraction(self.var.SeepSubToGW)
         # Pixel-average seepage values in [mm] per timestep
         # (no seepage from direct runoff fraction)
+        
+        # the variables below were added to report catchment-averaged soil moisture profiles
+        self.var.Theta1aPixel = self.var.deffraction(self.var.Theta1a)
+        self.var.Theta1bPixel = self.var.deffraction(self.var.Theta1b)
+        self.var.Theta2Pixel = self.var.deffraction(self.var.Theta2)
